@@ -1,6 +1,6 @@
 /*****************************************************************************
 Copyright (C) 2013, 2015, Google Inc. All Rights Reserved.
-Copyright (c) 2014, 2021, MariaDB Corporation.
+Copyright (c) 2014, 2020, MariaDB Corporation.
 
 This program is free software; you can redistribute it and/or modify it under
 the terms of the GNU General Public License as published by the Free Software
@@ -55,9 +55,6 @@ UNIV_INTERN uint srv_n_fil_crypt_threads_started = 0;
 /** At this age or older a space/page will be rotated */
 UNIV_INTERN uint srv_fil_crypt_rotate_key_age;
 
-/** Whether the encryption plugin does key rotation */
-static bool srv_encrypt_rotate;
-
 /** Event to signal FROM the key rotation threads. */
 static os_event_t fil_crypt_event;
 
@@ -139,14 +136,6 @@ fil_space_crypt_t::key_get_latest_version(void)
 
 	if (is_key_found()) {
 		key_version = encryption_key_get_latest_version(key_id);
-		/* InnoDB does dirty read of srv_fil_crypt_rotate_key_age.
-		It doesn't matter because srv_encrypt_rotate
-		can be set to true only once */
-		if (!srv_encrypt_rotate
-		    && key_version > srv_fil_crypt_rotate_key_age) {
-			srv_encrypt_rotate = true;
-		}
-
 		srv_stats.n_key_requests.inc();
 		key_found = key_version;
 	}
@@ -458,7 +447,6 @@ byte*
 fil_parse_write_crypt_data(
 	byte*			ptr,
 	const byte*		end_ptr,
-	const buf_block_t*	block,
 	dberr_t*		err)
 {
 	/* check that redo log entry is complete */
@@ -505,12 +493,12 @@ fil_parse_write_crypt_data(
 		return NULL;
 	}
 
-	mutex_enter(&fil_system->mutex);
+	mutex_enter(&fil_system.mutex);
 
 	fil_space_t* space = fil_space_get_by_id(space_id);
 
 	if (!space) {
-		mutex_exit(&fil_system->mutex);
+		mutex_exit(&fil_system.mutex);
 		return ptr + len;
 	}
 
@@ -531,7 +519,7 @@ fil_parse_write_crypt_data(
 		space->crypt_data = crypt_data;
 	}
 
-	mutex_exit(&fil_system->mutex);
+	mutex_exit(&fil_system.mutex);
 
 	if (crypt_data->should_encrypt() && !crypt_data->is_key_found()) {
 		*err = DB_DECRYPTION_FAILED;
@@ -592,7 +580,7 @@ fil_encrypt_buf(
 
 	int rc = encryption_scheme_encrypt(src, srclen, dst, &dstlen,
 					   crypt_data, key_version,
-					   space, offset, lsn);
+					   (uint32)space, (uint32)offset, lsn);
 	ut_a(rc == MY_AES_OK);
 	ut_a(dstlen == srclen);
 
@@ -657,7 +645,7 @@ fil_space_encrypt(
 		return (src_frame);
 	}
 
-	ut_ad(space->n_pending_ios > 0);
+	ut_ad(space->pending_io());
 	return fil_encrypt_buf(space->crypt_data, space->id, offset, lsn,
 			       src_frame, page_size_t(space->flags),
 			       dst_frame);
@@ -770,7 +758,7 @@ fil_space_decrypt(
 	const page_size_t page_size(space->flags);
 
 	ut_ad(space->crypt_data != NULL && space->crypt_data->is_encrypted());
-	ut_ad(space->n_pending_ios > 0);
+	ut_ad(space->pending_io());
 
 	bool encrypted = fil_space_decrypt(space->crypt_data, tmp_frame,
 					   page_size, src_frame, &err);
@@ -919,12 +907,12 @@ fil_crypt_read_crypt_data(fil_space_t* space)
 	mtr.start();
 	if (buf_block_t* block = buf_page_get(page_id_t(space->id, 0),
 					      page_size, RW_S_LATCH, &mtr)) {
-		mutex_enter(&fil_system->mutex);
+		mutex_enter(&fil_system.mutex);
 		if (!space->crypt_data) {
 			space->crypt_data = fil_space_read_crypt_data(
 				page_size, block->frame);
 		}
-		mutex_exit(&fil_system->mutex);
+		mutex_exit(&fil_system.mutex);
 	}
 	mtr.commit();
 }
@@ -977,9 +965,9 @@ static bool fil_crypt_start_encrypting_space(fil_space_t* space)
 	crypt_data->rotate_state.starting = true;
 	crypt_data->rotate_state.active_threads = 1;
 
-	mutex_enter(&fil_system->mutex);
+	mutex_enter(&fil_system.mutex);
 	space->crypt_data = crypt_data;
-	mutex_exit(&fil_system->mutex);
+	mutex_exit(&fil_system.mutex);
 
 	fil_crypt_start_converting = true;
 	mutex_exit(&fil_crypt_threads_mutex);
@@ -1012,11 +1000,13 @@ static bool fil_crypt_start_encrypting_space(fil_space_t* space)
 		/* 4 - sync tablespace before publishing crypt data */
 
 		bool success = false;
+		ulint sum_pages = 0;
 
 		do {
 			ulint n_pages = 0;
 			success = buf_flush_lists(ULINT_MAX, end_lsn, &n_pages);
 			buf_flush_wait_batch_end(NULL, BUF_FLUSH_LIST);
+			sum_pages += n_pages;
 		} while (!success);
 
 		/* 5 - publish crypt data */
@@ -1092,33 +1082,6 @@ struct rotate_thread_t {
 	}
 };
 
-/** Avoid the removal of the tablespace from
-default_encrypt_list only when
-1) Another active encryption thread working on tablespace
-2) Eligible for tablespace key rotation
-3) Tablespace is in flushing phase
-@return true if tablespace should be removed from
-default encrypt */
-static bool fil_crypt_must_remove(const fil_space_t &space)
-{
-  ut_ad(space.purpose == FIL_TYPE_TABLESPACE);
-  fil_space_crypt_t *crypt_data = space.crypt_data;
-  ut_ad(mutex_own(&fil_system->mutex));
-  const ulong encrypt_tables= srv_encrypt_tables;
-  if (!crypt_data)
-    return !encrypt_tables;
-  if (!crypt_data->is_key_found())
-    return true;
-
-  mutex_enter(&crypt_data->mutex);
-  const bool remove= (space.is_stopping() || crypt_data->not_encrypted()) &&
-    (!crypt_data->rotate_state.flushing &&
-     !encrypt_tables == !!crypt_data->min_key_version &&
-     !crypt_data->rotate_state.active_threads);
-  mutex_exit(&crypt_data->mutex);
-  return remove;
-}
-
 /***********************************************************************
 Check if space needs rotation given a key_state
 @param[in,out]		state		Key rotation state
@@ -1139,7 +1102,7 @@ fil_crypt_space_needs_rotation(
 		return false;
 	}
 
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 
 	fil_space_crypt_t *crypt_data = space->crypt_data;
 
@@ -1200,7 +1163,7 @@ fil_crypt_space_needs_rotation(
 			key_state->rotate_key_age);
 
 		crypt_data->rotate_state.scrubbing.is_active =
-			btr_scrub_start_space(*space, &state->scrub_data);
+			btr_scrub_start_space(space->id, &state->scrub_data);
 
 		time_t diff = time(0) - crypt_data->rotate_state.scrubbing.
 			last_scrub_completed;
@@ -1416,70 +1379,56 @@ fil_crypt_return_iops(
 	fil_crypt_update_total_stat(state);
 }
 
-bool fil_crypt_must_default_encrypt()
-{
-  return !srv_fil_crypt_rotate_key_age || !srv_encrypt_rotate;
-}
-
-/** Return the next tablespace from default_encrypt_tables.
+/** Return the next tablespace from rotation_list.
 @param space   previous tablespace (NULL to start from the start)
 @param recheck whether the removal condition needs to be rechecked after
 the encryption parameters were changed
 @param encrypt expected state of innodb_encrypt_tables
 @return the next tablespace to process (n_pending_ops incremented)
 @retval NULL if this was the last */
-inline fil_space_t *fil_system_t::default_encrypt_next(
-  fil_space_t *space, bool recheck, bool encrypt)
+inline fil_space_t *fil_system_t::keyrotate_next(fil_space_t *space,
+                                                 bool recheck, bool encrypt)
 {
   ut_ad(mutex_own(&mutex));
 
   sized_ilist<fil_space_t, rotation_list_tag_t>::iterator it=
-    space && space->is_in_default_encrypt
-    ? space
-    : default_encrypt_tables.begin();
+    space && space->is_in_rotation_list ? space : rotation_list.begin();
   const sized_ilist<fil_space_t, rotation_list_tag_t>::iterator end=
-    default_encrypt_tables.end();
+    rotation_list.end();
 
   if (space)
   {
-    const bool released= !--space->n_pending_ops;
+    const bool released= !space->release();
 
-    if (space->is_in_default_encrypt)
+    if (space->is_in_rotation_list)
     {
       while (++it != end &&
              (!UT_LIST_GET_LEN(it->chain) || it->is_stopping()));
 
-      /* If one of the encryption threads already started
-      the encryption of the table then don't remove the
-      unencrypted spaces from default encrypt list.
+      /* If one of the encryption threads already started the encryption
+      of the table then don't remove the unencrypted spaces from rotation list
 
-      If there is a change in innodb_encrypt_tables variables
-      value then don't remove the last processed tablespace
-      from the default encrypt list. */
-      if (released && !recheck && fil_crypt_must_remove(*space))
+      If there is a change in innodb_encrypt_tables variables value then
+      don't remove the last processed tablespace from the rotation list. */
+      if (released && (!recheck || space->crypt_data) &&
+          !encrypt == !srv_encrypt_tables)
       {
-        ut_a(!default_encrypt_tables.empty());
-        default_encrypt_tables.remove(*space);
-        space->is_in_default_encrypt= false;
+        ut_a(!rotation_list.empty());
+        rotation_list.remove(*space);
+        space->is_in_rotation_list= false;
       }
     }
   }
-  else while (it != end &&
-	      (!UT_LIST_GET_LEN(it->chain) || it->is_stopping()))
+
+  while (it != end)
   {
-    /* Find the next suitable default encrypt table if
-    beginning of default_encrypt_tables list has been scheduled
-    to be deleted */
-    it++;
+    space= &*it;
+    if (space->acquire())
+      return space;
+    while (++it != end && (!UT_LIST_GET_LEN(it->chain) || it->is_stopping()));
   }
 
-  if (it == end)
-    return NULL;
-
-  space= &*it;
-  space->n_pending_ops++;
-  ut_ad(!space->is_stopping());
-  return space;
+  return NULL;
 }
 
 /** Return the next tablespace.
@@ -1492,23 +1441,23 @@ the encryption parameters were changed
 static fil_space_t *fil_space_next(fil_space_t *space, bool recheck,
                                    bool encrypt)
 {
-  mutex_enter(&fil_system->mutex);
-  ut_ad(!space || space->n_pending_ops);
+  mutex_enter(&fil_system.mutex);
 
-  if (fil_crypt_must_default_encrypt())
-    space= fil_system->default_encrypt_next(space, recheck, encrypt);
+  if (!srv_fil_crypt_rotate_key_age)
+    space= fil_system.keyrotate_next(space, recheck, encrypt);
   else if (!space)
   {
-    space= UT_LIST_GET_FIRST(fil_system->space_list);
+    space= UT_LIST_GET_FIRST(fil_system.space_list);
     /* We can trust that space is not NULL because at least the
     system tablespace is always present and loaded first. */
-    space->n_pending_ops++;
+    if (!space->acquire())
+      goto next;
   }
   else
   {
-    ut_ad(space->n_pending_ops > 0);
     /* Move on to the next fil_space_t */
-    space->n_pending_ops--;
+    space->release();
+next:
     space= UT_LIST_GET_NEXT(space_list, space);
 
     /* Skip abnormal tablespaces or those that are being created by
@@ -1518,11 +1467,11 @@ static fil_space_t *fil_space_next(fil_space_t *space, bool recheck,
             space->is_stopping() || space->purpose != FIL_TYPE_TABLESPACE))
       space= UT_LIST_GET_NEXT(space_list, space);
 
-    if (space)
-      space->n_pending_ops++;
+    if (space && !space->acquire())
+      goto next;
   }
 
-  mutex_exit(&fil_system->mutex);
+  mutex_exit(&fil_system.mutex);
   return space;
 }
 
@@ -1539,7 +1488,7 @@ static bool fil_crypt_find_space_to_rotate(
 	/* we need iops to start rotating */
 	while (!state->should_shutdown() && !fil_crypt_alloc_iops(state)) {
 		if (state->space && state->space->is_stopping()) {
-			fil_space_release(state->space);
+			state->space->release();
 			state->space = NULL;
 		}
 
@@ -1549,7 +1498,7 @@ static bool fil_crypt_find_space_to_rotate(
 
 	if (state->should_shutdown()) {
 		if (state->space) {
-			fil_space_release(state->space);
+			state->space->release();
 			state->space = NULL;
 		}
 		return false;
@@ -1558,7 +1507,7 @@ static bool fil_crypt_find_space_to_rotate(
 	if (state->first) {
 		state->first = false;
 		if (state->space) {
-			fil_space_release(state->space);
+			state->space->release();
 		}
 		state->space = NULL;
 	}
@@ -1587,7 +1536,7 @@ static bool fil_crypt_find_space_to_rotate(
 	}
 
 	if (state->space) {
-		fil_space_release(state->space);
+		state->space->release();
 		state->space = NULL;
 	}
 
@@ -1661,7 +1610,7 @@ fil_crypt_find_page_to_rotate(
 	ulint batch = srv_alloc_time * state->allocated_iops;
 	fil_space_t* space = state->space;
 
-	ut_ad(!space || space->n_pending_ops > 0);
+	ut_ad(!space || space->referenced());
 
 	/* If space is marked to be dropped stop rotation. */
 	if (!space || space->is_stopping()) {
@@ -1719,7 +1668,7 @@ fil_crypt_get_page_throttle_func(
 	fil_space_t* space = state->space;
 	const page_size_t page_size = page_size_t(space->flags);
 	const page_id_t page_id(space->id, offset);
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 
 	/* Before reading from tablespace we need to make sure that
 	the tablespace is not about to be dropped. */
@@ -1801,11 +1750,11 @@ btr_scrub_get_block_and_allocation_status(
 	buf_block_t *block = NULL;
 	fil_space_t* space = state->space;
 
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 
 	mtr_start(&local_mtr);
 
-	*allocation_status = fseg_page_is_free(space, offset) ?
+	*allocation_status = fseg_page_is_free(space, (uint32_t)offset) ?
 		BTR_SCRUB_PAGE_FREE :
 		BTR_SCRUB_PAGE_ALLOCATED;
 
@@ -1853,7 +1802,7 @@ fil_crypt_rotate_page(
 	ulint sleeptime_ms = 0;
 	fil_space_crypt_t *crypt_data = space->crypt_data;
 
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 	ut_ad(offset > 0);
 
 	/* In fil_crypt_thread where key rotation is done we have
@@ -2040,7 +1989,7 @@ fil_crypt_rotate_pages(
 	ulint end = std::min(state->offset + state->batch,
 			     state->space->free_limit);
 
-	ut_ad(state->space->n_pending_ops > 0);
+	ut_ad(state->space->referenced());
 
 	for (; state->offset < end; state->offset++) {
 
@@ -2079,7 +2028,7 @@ fil_crypt_flush_space(
 	fil_space_t* space = state->space;
 	fil_space_crypt_t *crypt_data = space->crypt_data;
 
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 
 	/* flush tablespace pages so that there are no pages left with old key */
 	lsn_t end_lsn = crypt_data->rotate_state.end_lsn;
@@ -2134,18 +2083,13 @@ fil_crypt_flush_space(
 
 /***********************************************************************
 Complete rotating a space
-@param[in,out]		key_state		Key state
 @param[in,out]		state			Rotation state */
-static
-void
-fil_crypt_complete_rotate_space(
-	const key_state_t*	key_state,
-	rotate_thread_t*	state)
+static void fil_crypt_complete_rotate_space(rotate_thread_t* state)
 {
 	fil_space_crypt_t *crypt_data = state->space->crypt_data;
 
 	ut_ad(crypt_data);
-	ut_ad(state->space->n_pending_ops > 0);
+	ut_ad(state->space->referenced());
 
 	/* Space might already be dropped */
 	if (!state->space->is_stopping()) {
@@ -2298,9 +2242,8 @@ DECLARE_THREAD(fil_crypt_thread)(void*)
 				/* If space is marked as stopping, release
 				space and stop rotation. */
 				if (thr.space->is_stopping()) {
-					fil_crypt_complete_rotate_space(
-						&new_state, &thr);
-					fil_space_release(thr.space);
+					fil_crypt_complete_rotate_space(&thr);
+					thr.space->release();
 					thr.space = NULL;
 					break;
 				}
@@ -2311,7 +2254,7 @@ DECLARE_THREAD(fil_crypt_thread)(void*)
 
 			/* complete rotation */
 			if (thr.space) {
-				fil_crypt_complete_rotate_space(&new_state, &thr);
+				fil_crypt_complete_rotate_space(&thr);
 			}
 
 			/* force key state refresh */
@@ -2327,7 +2270,7 @@ DECLARE_THREAD(fil_crypt_thread)(void*)
 
 	/* release current space if shutting down */
 	if (thr.space) {
-		fil_space_release(thr.space);
+		thr.space->release();
 		thr.space = NULL;
 	}
 
@@ -2388,42 +2331,38 @@ fil_crypt_set_thread_cnt(
 	}
 }
 
-/** Initialize the tablespace default_encrypt_tables
+/** Initialize the tablespace rotation_list
 if innodb_encryption_rotate_key_age=0. */
-static void fil_crypt_default_encrypt_tables_fill()
+static void fil_crypt_rotation_list_fill()
 {
-	ut_ad(mutex_own(&fil_system->mutex));
+	ut_ad(mutex_own(&fil_system.mutex));
 
-	for (fil_space_t* space = UT_LIST_GET_FIRST(fil_system->space_list);
+	for (fil_space_t* space = UT_LIST_GET_FIRST(fil_system.space_list);
 	     space != NULL;
 	     space = UT_LIST_GET_NEXT(space_list, space)) {
 		if (space->purpose != FIL_TYPE_TABLESPACE
-		    || space->is_in_default_encrypt
-		    || space->is_stopping()
-		    || UT_LIST_GET_LEN(space->chain) == 0) {
+		    || space->is_in_rotation_list
+		    || UT_LIST_GET_LEN(space->chain) == 0
+		    || !space->acquire()) {
 			continue;
 		}
 
 		/* Ensure that crypt_data has been initialized. */
 		if (!space->size) {
-			/* Protect the tablespace while we may
-			release fil_system->mutex. */
-			space->n_pending_ops++;
 			ut_d(const fil_space_t* s=)
-				fil_system->read_page0(space->id);
+			        fil_system.read_page0(space->id);
 			ut_ad(!s || s == space);
-			space->n_pending_ops--;
 			if (!space->size) {
 				/* Page 0 was not loaded.
 				Skip this tablespace. */
-				continue;
+				goto next;
 			}
 		}
 
 		/* Skip ENCRYPTION!=DEFAULT tablespaces. */
 		if (space->crypt_data
 		    && !space->crypt_data->is_default_encryption()) {
-			continue;
+			goto next;
 		}
 
 		if (srv_encrypt_tables) {
@@ -2431,19 +2370,21 @@ static void fil_crypt_default_encrypt_tables_fill()
 			innodb_encrypt_tables!=OFF */
 			if (space->crypt_data
 			    && space->crypt_data->min_key_version) {
-				continue;
+				goto next;
 			}
 		} else {
 			/* Skip unencrypted tablespaces if
 			innodb_encrypt_tables=OFF */
 			if (!space->crypt_data
 			    || !space->crypt_data->min_key_version) {
-				continue;
+				goto next;
 			}
 		}
 
-		fil_system->default_encrypt_tables.push_back(*space);
-		space->is_in_default_encrypt = true;
+		fil_system.rotation_list.push_back(*space);
+		space->is_in_rotation_list = true;
+next:
+		space->release();
 	}
 }
 
@@ -2455,12 +2396,12 @@ void
 fil_crypt_set_rotate_key_age(
 	uint	val)
 {
-	mutex_enter(&fil_system->mutex);
+	mutex_enter(&fil_system.mutex);
 	srv_fil_crypt_rotate_key_age = val;
 	if (val == 0) {
-		fil_crypt_default_encrypt_tables_fill();
+		fil_crypt_rotation_list_fill();
 	}
-	mutex_exit(&fil_system->mutex);
+	mutex_exit(&fil_system.mutex);
 	os_event_set(fil_crypt_threads_event);
 }
 
@@ -2484,15 +2425,15 @@ void
 fil_crypt_set_encrypt_tables(
 	uint val)
 {
-	mutex_enter(&fil_system->mutex);
+	mutex_enter(&fil_system.mutex);
 
 	srv_encrypt_tables = val;
 
-	if (fil_crypt_must_default_encrypt()) {
-		fil_crypt_default_encrypt_tables_fill();
+	if (srv_fil_crypt_rotate_key_age == 0) {
+		fil_crypt_rotation_list_fill();
 	}
 
-	mutex_exit(&fil_system->mutex);
+	mutex_exit(&fil_system.mutex);
 
 	os_event_set(fil_crypt_threads_event);
 }
@@ -2603,7 +2544,7 @@ fil_space_crypt_get_status(
 {
 	memset(status, 0, sizeof(*status));
 
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 
 	/* If there is no crypt data and we have not yet read
 	page 0 for this tablespace, we need to read it before
@@ -2668,7 +2609,7 @@ fil_space_get_scrub_status(
 {
 	memset(status, 0, sizeof(*status));
 
-	ut_ad(space->n_pending_ops > 0);
+	ut_ad(space->referenced());
 	fil_space_crypt_t* crypt_data = space->crypt_data;
 
 	status->space = space->id;
